gstrGameVersion = "0.01"

inspect = require 'inspect'
-- https://github.com/kikito/inspect.lua

TLfres = require "tlfres"
-- https://love2d.org/wiki/TLfres

ft = require "foxtree"
bt = require "behaviortree"
cobjs = require "createobjects"
dobjs = require "drawobjects"
fun = require "randomfunctions"
enum = require "enum"
cf = require "commonfunctions"

gintScreenWidth = 1440-- 1920
gintScreenHeight = 900-- 1080
garrCurrentScreen = {}			-- screen stack

gintAgentRadius = 5
gintTransactions = 0
gintDeaths = 0
gintCurrentCases = 0
gintTotalWealth = 0

gfltTotalTime = 0				-- sum of all dt's for this session


Agents = {}						-- these are physics objects
Zones = {}						-- area's of interest to agents
tree = {}


gbolDebug = true
gbolPaused = false
gbolShowData = false

gtmrDegradeStats = enum.timerDegradeStats
gtmrGetStats = enum.timerGetStats
gtmrSpawnAgents = enum.timerSpawnAgents
gtmrKillThings = enum.timerKillThings

HospitalX = 1
HospitalY = 600
HospitalWidth = 100
HospitalHeight = 300

IsoX = 1
IsoY = 300
IsoWidth = 100
IsoHeight = 300


local function MoveAgent(v)
-- moves a single bot towards tx,ty
-- v = a single agent

	if v.targetx ~= nil and v.targety ~= nill then
		-- determine vector to get to next waypoint
		local xdistance = v.targetx - v.body:getX()		-- x axis set above
		local ydistance = v.targety - v.body:getY()		-- y axis
		
		local xvector = xdistance
		local yvector = ydistance

		-- need to scale to 'walking' pace
		xvector, yvector = fun.NormaliseVectors(xvector, yvector)
		
		v.body:setLinearVelocity(xvector, yvector)
	end
end

local function UpdateTaskTimer(v, dt)
-- v = single bot
	v.nexttasktimer = v.nexttasktimer - dt
	if v.nexttasktimer <= 0 then 
		-- set up for the next task
		v.nexttasktimer = 0
		v.targetx = nil
		v.targety = nil
		v.currenttask = nil
		v.targetzone = nil
	end	
end

local function EnsureTargetSet(z, v,targetzonetype)
-- v = single agent
-- targetzonetype = the type of zone to look for

	if v.targetx == nil then
		-- find a zone
		local rndnum
		repeat
			rndnum = love.math.random(1, #z)
		until z.zonetype == targetzonetype

		z[rndnum].
		v.targetx = z[rndnum].x + (z[rndnum].width / 2)
		v.targety = z[rndnum].y + (z[rndnum].height / 2)
		v.targetzone = q		
	end	
end

local function StopIfAtTarget(v)
-- returns a boolean AND stops the agent moving
-- v = single agent

	if cf.GetDistance(v.body:getX(), v.body:getY(),v.targetx,v.targety) < 25 then
		-- arrived. Remember we have arrived
		v.nexttasktimer = enum.timerNextTask
		local vx, vy = v.body:getLinearVelocity( )
		v.body:setLinearVelocity(vx / 2, vy / 2)
		return true
	end
	return false

end

local function CheckIdleAgents(agt,dt)
-- agt = Agents

	for k,v in ipairs(agt) do
		v.nexttasktimer = v.nexttasktimer - dt
		if v.nexttasktimer <= 0 then v.nexttasktimer = 0 end

		-- see who is idle
		if v.currenttask == nil and v.nexttasktimer <= 0 then
			nextaction = ft.DetermineAction(tree, v)
			
			v.currenttask = nextaction
			
			if nextaction == enum.goalRest then
				v.currenttasklabel = "Resting"
			end
			if nextaction == enum.goalWork then
				v.currenttasklabel = "Working"
			end
			if nextaction == enum.goalHeal then
				v.currenttasklabel = "Needs doctor"
			end		
			if nextaction == enum.goalHomeIso then
				v.currenttasklabel = "Isolation"
			end	
		end
	end
end

function DegradeStats(agt, dt)
-- adjust vital stats every tick
-- agt = Agents

	gfltTotalTime = gfltTotalTime + dt
	
	gtmrDegradeStats = gtmrDegradeStats - dt
	if gtmrDegradeStats <= 0 then
		gtmrDegradeStats = enum.timerDegradeStats
		
		for k,v in ipairs(agt) do
			if v.currenttask == goalHomeIso then
				v.health = v.health + 3			-- natural ability to recover if resting in iso. Needed to escape iso
			end

			if v.viralload > 0 then
				v.health = v.health - 1
			end
			if v.viralload > 0 and v.maxviralload > 0 then
				v.viralload = v.viralload + 1
				if v.viralload >= v.maxviralload then
					v.maxviralload = 0
				end
			end
			if v.viralload > 0 and v.maxviralload == 0 then
				v.viralload = v.viralload - 1
				if v.viralload < 1 and v.virusdetected and currenttask == enum.goalHeal then
					gintCurrentCases = gintCurrentCases - 1
--gbolPaused = true
print("alpha")
				end
			end
			if v.viralload <= 0 then
				v.shedding = 0
			end
			
		end
	end
end

function CheckStatBounds(agt)
-- simple check for zero's and 100's

	for k,v in ipairs(agt) do
		if v.health >= 100 then v.health = 100 end
		if v.happiness >= 100 then v.happiness = 100 end

		if v.health <= 0 then v.health = 0 end
		if v.happiness <= 0 then v.happiness = 0 end
		if v.wealth <= 0 then v.wealth = 0 end
	
	end
end

function GetStats(agt, dt)

	gtmrGetStats = gtmrGetStats - dt
	if gtmrGetStats <= 0 then
		gtmrGetStats = enum.timerGetStats
		
		gintTotalWealth = 0
		for k,v in ipairs(agt) do
			gintTotalWealth = gintTotalWealth + v.wealth
		end
	end

end

local function PerformRest(v, dt)
-- v = single agent

	if v.nexttasktimer <= 0 then	-- we have a task but the task timer is not yet zero
		-- remember the target if necessary
		if v.targetx == nil then
			-- pick a random spot nearby

			local x = v.body:getX()
			local y = v.body:getY()			

			v.targetx = x + love.math.random (-150, 150)
			v.targety = y + love.math.random (-150, 150)
			v.targetzone = nil
			
			if v.targetx < HospitalX + HospitalWidth + 150 then v.targetx = HospitalX + HospitalWidth + 150 end
			
			if v.job == enum.jobTester then
				v.red = 226/255
				v.green = 75/255
				v.blue = 211/255			
			else
				v.red = 1
				v.green = 1
				v.blue = 1			
			end
			v.fixture:setSensor(false)		-- false = bounce	
		end	
		-- move towards target
		MoveAgent(v)
		
		if StopIfAtTarget(v) then
			-- rest

		end	
	else
		-- arrived some time previously
		UpdateTaskTimer(v, dt)
	end

end

local function PerformWork(v, dt)
-- v = single agent

	if v.nexttasktimer <= 0 then	-- we have a task but the task timer is not yet zero
		-- remember the target if necessary
		if v.targetx == nil then
			-- pick a random spot nearby

			local x = v.body:getX()
			local y = v.body:getY()			

			v.targetx = x + love.math.random (-400, 400)
			v.targety = y + love.math.random (-400, 400)
			
			if v.targetx < 300 then v.targetx = 300 end
			if v.targetx > 1200 then v.targetx = 1200 end
			if v.targety < 300 then v.targety = 300 end
			if v.targety > 700 then v.targety = 700 end
			
			v.targetzone = nil
			
			if v.job == enum.jobTester then
				v.red = 226/255
				v.green = 75/255
				v.blue = 211/255			
			else			
				v.red = 1
				v.green = 1
				v.blue = 1			
			end
			v.fixture:setSensor(false)	-- false = bounce			
		end	
		-- move towards target
		MoveAgent(v)
		
		if StopIfAtTarget(v) then
			-- Earn money
			v.wealth = v.wealth + 1
		end	
	else
		-- arrived some time previously
		UpdateTaskTimer(v, dt)
	end

end

local function PerformHeal(v, dt)
-- v = single agent

	if v.nexttasktimer <= 0 then	-- we have a task but the task timer is not yet zero
		-- remember the target if necessary
		if v.targetx == nil then
			-- pick a random spot nearby

		
			v.targetx = HospitalX + love.math.random (0, HospitalWidth)
			v.targety = HospitalY + love.math.random (0, HospitalHeight)
	
			if v.targetx < HospitalX then v.targetx = HospitalX end
			if v.targetx > gintScreenWidth then v.targetx = gintScreenWidth end
			if v.targety < HospitalY then v.targety = HospitalY end
			if v.targety > gintScreenHeight then v.targety = gintScreenHeight end

			v.targetzone = nil
			
			v.red = 1
			v.green = 0
			v.blue = 0			
			
			v.fixture:setSensor(true)	-- false = bounce		
		end	
		-- move towards target
		MoveAgent(v)
		
		if StopIfAtTarget(v) then
			-- heal
			v.health = v.health + 0.7
			v.wealth = v.wealth - 10		-- hospital has passive and direct costs
		
			if v.viralload > 0 and v.virusdetected == false then
				v.virusdetected = true
				gintCurrentCases = gintCurrentCases + 1
				-- gbolPaused = true
			end
			
		end	
	else 
		-- arrived some time previously
		UpdateTaskTimer(v, dt)
		-- clear virus flag if finished ISO
		if v.nexttasktimer <= 0 and v.virusdetected then
			-- retest the bot for the virus
			if v.viralload <= 0 then
				v.virusdetected = false
				gintCurrentCases = gintCurrentCases - 1
--gbolPaused = true
print("beta")
			else
				-- still infected but let the tree decide if they move to iso
				v.virusdetected = true
			end
		end		
	end


end

local function PerformHomeIso(v, dt)

	if v.nexttasktimer <= 0 then	-- we have a task but the task timer is not yet zero
		-- remember the target if necessary
		if v.targetx == nil then
			-- pick a random spot nearby
			v.targetx = IsoX + love.math.random (0, IsoWidth)
			v.targety = IsoY + love.math.random (0, IsoHeight)
	
			if v.targetx < IsoX then v.targetx = IsoX end
			if v.targetx > gintScreenWidth then v.targetx = gintScreenWidth end
			if v.targety < IsoY then v.targety = IsoY end
			if v.targety > gintScreenHeight then v.targety = gintScreenHeight end

			v.targetzone = nil
			
			v.red = 142/255
			v.green = 139/255
			v.blue = 240/255			
			
			v.fixture:setSensor(true)		-- false = bounce	
		end	
		-- move towards target
		MoveAgent(v)
		
		if StopIfAtTarget(v) then
			-- make isolation twice the normal timer
			v.health = v.health + 0.7
			v.timerNextTask = enum.timerNextTask * 2
		end	
	else
		-- arrived some time previously
		UpdateTaskTimer(v, dt)
		-- clear virus flag if finished ISO and a negative test is returned
		if v.nexttasktimer <= 0 and v.viralload <= 0 then
			v.virusdetected = false
			gintCurrentCases = gintCurrentCases - 1
--gbolPaused = true
print("charlie")
		else
			-- need more iso
			v.timerNextTask = enum.timerNextTask * 2
		end
	end
end

local function PerformTasks(znes, agt, dt)
-- do things
-- zne = Zones

	for k,v in ipairs(agt) do
		if v.currenttask == enum.goalRest then
			PerformRest(v,dt)
		end	
		if v.currenttask == enum.goalWork then
			PerformWork(v,dt)
		end		
		if v.currenttask == enum.goalHeal then
			PerformHeal(v,dt)
		end	
		if v.currenttask == enum.goalHomeIso then
			PerformHomeIso(v,dt)
		end	
		
	end
end

local function SpawnAgents(dt)
-- spawn a new agent if conditions are good

	gtmrSpawnAgents = gtmrSpawnAgents - dt
	
	if gtmrSpawnAgents <= 0 then
		gtmrSpawnAgents = enum.timerSpawnAgents
	
		--!do things

	end
end

local function KillThings(Agts, dt)
	gtmrKillThings = gtmrKillThings - dt
	
	if gtmrKillThings <= 0 then
		gtmrKillThings = enum.timerKillThings
		
		for k,v in ipairs(Agts) do
			if v.health <= 0 then
				-- ack!
				local deadID = v.ID
				if v.virusdetected then
					gintCurrentCases = gintCurrentCases - 1
--gbolPaused = true
print("delta")
				end
	
				table.remove(Agts, k)
				gintDeaths = gintDeaths + 1
			end
		end
	end
end
 
function beginContact(a, b, coll)

	local bot1
	local bot2
	
	local bot1ID = a:getUserData()
	local bot2ID = b:getUserData()
	

	-- translate the bot.ID to the Agents index
	for k,v in ipairs(Agents) do
		if v.ID == bot1ID then
			bot1 = v			-- the whole bot
		end
		if v.ID == bot2ID then
			bot2 = v
		end		
	end

	-- I think some bots die as they contact (random luck) and that makes this bomb out
	if bot1 ~= nil and bot2 ~= nil then
	
		local x1 = bot1.body:getX()
		local y1 = bot1.body:getY()	
		local x2 = bot2.body:getX()
		local y2 = bot2.body:getY()		
	
		if bot1.wealth > 0 then
			-- check if in hospital
			if x1 <= HospitalX + HospitalWidth and y1 > HospitalY  then
				-- dude is in hospital. Do no transation 
			elseif x2 <= HospitalX + HospitalWidth and y2 > HospitalY  then
				-- dude is in hospital. Do no transation 
			else
				bot1.wealth = bot1.wealth - 1
				bot2.wealth = bot2.wealth + 1
				gintTransactions = gintTransactions + 1
			end
		end
		
		-- see if pass infection
		-- can't get infected in iso because - iso
		
		if x1 < IsoX + IsoWidth and y1 > IsoY and y1 < IsoY + IsoHeight then
			-- in Iso - can't get infected
		else
			if bot1.viralload <= 0 then
				local bolInfected = false
				if bot1.job == enum.jobTester then
					bolInfected = love.math.random(0,100) < bot2.shedding / 2
				else
					bolInfected = love.math.random(0,100) < bot2.shedding
				end	
				if bolInfected then
					-- infected
					bot1.viralload = 1
					bot1.maxviralload = love.math.random(3,125)
					bot1.shedding = love.math.random(15,50)
				end
			end
		end
			
		if x2 < IsoX + IsoWidth and y2 > IsoY and y2 < IsoY + IsoHeight then	
			-- in Iso - can't get infected
		else
			if bot2.viralload <= 0 then
				local bolInfected = false
				if bot2.job == enum.jobTester then
					bolInfected = love.math.random(0,100) < bot1.shedding / 2
				else
					bolInfected = love.math.random(0,100) < bot1.shedding
				end	
				if bolInfected then			
					if love.math.random(0, 100) < bot1.shedding then
						-- infected
						bot2.viralload = 1
						bot2.maxviralload = love.math.random(3,150)
						bot2.shedding = love.math.random(15,50)
					end
				end
			end	
		end
	
		-- see if a tester detects a virus
		if bot1.job == enum.jobTester or bot2.job == enum.jobTester then
			if bot1.viralload > 0 then 
				bot1.virusdetected = true
			end
			if bot2.viralload > 0 then
				bot2.virusdetected = true
			end
		end
	
	end
end

function endContact(a, b, coll)

	local bot1ID = a:getUserData()
	local bot2ID = b:getUserData()

end

function love.mousepressed( x, y, button, istouch, presses )

	-- local mousex,mousey = TLfres.getMousePosition(gintScreenWidth, gintScreenHeight)    -- lets you pretend screen is 1920 * 1080

	-- if button == 1 then
		-- for k,v in ipairs(Agents) do
			-- local x2 = v.body:getX()		-- x axis set above
			-- local y2 = v.body:getY()		-- y axis
			-- local dist = cf.GetDistance(mousex, mousey, x2, y2)

			-- if dist <= gintAgentRadius and v.occupation == nil then
				-- v.isselected = not v.isselected
			-- end
		-- end
	-- end
end

function love.keyreleased( key, scancode )
	if key == "escape" then
		fun.RemoveScreen()
	end
	if key == "space" then
		gbolPaused = not gbolPaused
	end
	
	if key == "lalt" or key == "ralt" then
		gbolShowData = not gbolShowData
	end
	
end

function love.load()

    if love.filesystem.isFused( ) then
        void = love.window.setMode(gintScreenWidth, gintScreenHeight,{fullscreen=false,display=1,resizable=true, borderless=false})	-- display = monitor number (1 or 2)
        gbolDebug = false
    else
        void = love.window.setMode(gintScreenWidth, gintScreenHeight,{fullscreen=false,display=2,resizable=true, borderless=false})	-- display = monitor number (1 or 2)
    end
	
	love.window.setTitle("Lockdown " .. gstrGameVersion)
	
	love.physics.setMeter(1)
	world = love.physics.newWorld(0,0,false)	
	world:setCallbacks(beginContact,_,_,_)
	
	cobjs.CreateWorld(Zones)
	--cobjs.CreateWorld(Zones)
	gintNumAgents = 1000 --200
	for i = 1,gintNumAgents do
		cobjs.CreateAgent(i) 
		
		if i <= 5 then
			Agents[i].viralload = 50
			Agents[i].maxviralload = 60
			Agents[i].shedding = 50
		end
		if i > 5 and i < 101 then
			Agents[i].job = enum.jobTester
			Agents[i].red = 226/255
			Agents[i].green = 75/255
			Agents[i].blue = 211/255		
		
		end
		
		
	end
	


	fun.AddScreen("World")

	bt.EstablishTree()
	
	cf.beep()
    
end

function love.draw()

	
	TLfres.beginRendering(gintScreenWidth,gintScreenHeight)

	dobjs.DrawWorld()
	dobjs.DrawAgents(Agents)
	
	dobjs.DrawInstructions()
	
	dobjs.DrawStats()
	
	if not love.filesystem.isFused( ) then
		
		
		-- dobjs.DrawPriorities()
	end
	TLfres.endRendering({0, 0, 0, 1})

end

function love.update(dt)

	if not gbolPaused then
		CheckIdleAgents(Agents,dt)
		PerformTasks(Zones, Agents, dt)
		CheckStatBounds(Agents)

		DegradeStats(Agents, dt)
		KillThings(Agents, dt)
		
		GetStats(Agents, dt)
		
		SpawnAgents(dt)
		
		--BalancePriorities(Agents, dt)
		world:update(dt) --this puts the world into motion
	end
end







